{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "import io\n",
    "import os\n",
    "import ast\n",
    "import csv\n",
    "import math\n",
    "import keyword\n",
    "import tokenize\n",
    "import statistics\n",
    "from io import BytesIO\n",
    "from radon.raw import analyze\n",
    "from collections import deque\n",
    "from collections import Counter\n",
    "from radon.metrics import h_visit, ComplexityVisitor, mi_visit"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [],
   "source": [
    "#logger\n",
    "class Logger:\n",
    "    log_file_path = \"../../logs/log.log\" # or any other path\n",
    "\n",
    "    if not os.path.exists(\"../../logs\"):\n",
    "        os.makedirs(\"../../logs\")\n",
    "\n",
    "    @classmethod\n",
    "    def error_log(cls, s: str):\n",
    "        with open(cls.log_file_path, \"a\") as error_f:\n",
    "            error_f.write(f\"Error: {s}\\n\")\n",
    "\n",
    "    @classmethod\n",
    "    def info_log(cls, s: str):\n",
    "        with open(cls.log_file_path, \"a\") as info_f:\n",
    "            info_f.write(f\"Info: {s}\\n\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [],
   "source": [
    "nan = math.nan\n",
    "default_values = [nan, -1, -1.0]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [],
   "source": [
    "class CodeMetricsExtractor:\n",
    "    def __init__(self, code):\n",
    "        # library-independent\n",
    "        self.code: str = code\n",
    "\n",
    "        self.line_length = self._get_line_length()\n",
    "        self.whitespaces = self._get_num_whitespaces()\n",
    "        self.empty_lines = self._get_empty_lines()\n",
    "        self.code_length = len(code)\n",
    "        \n",
    "\n",
    "        # ast-dependent\n",
    "        self.tree = self._get_ast()\n",
    "        self.num_functions = self._get_num_functions()\n",
    "        self.functions_params = self._get_functions_params()\n",
    "\n",
    "        # tokenize-dependent\n",
    "        self.tokens = self._get_tokens()\n",
    "        self.keywords = self._get_keywords()\n",
    "\n",
    "        self.radon_raw_metrics = self._get_radon_metrics()\n",
    "\n",
    "        self.metrics = {}\n",
    "        if self.code_length == 0 or not self.code.strip():\n",
    "            Logger.error_log(\"File is empty or contains only whitespaces\")\n",
    "        elif self.tree is None:\n",
    "            Logger.error_log(\"Failed to parse code to ast.\")\n",
    "        elif self.tokens is None:\n",
    "            Logger.error_log(\"Failed to tokenize code.\")\n",
    "        elif self.radon_raw_metrics is None:\n",
    "            Logger.error_log(\"Failed to calculate Radon metrics.\")\n",
    "        else:\n",
    "            self._get_library_independent_metrics()\n",
    "            self._get_ast_dependent_metrics()\n",
    "            self._get_tokenize_dependent_metrics()\n",
    "            self._get_radon_dependent_metrics()\n",
    "\n",
    "    def _get_library_independent_metrics(self):\n",
    "        \"\"\"\n",
    "        Returns metrics that do not require importing any library/module\n",
    "        \"\"\"\n",
    "        self.metrics.update({ \n",
    "            'avgLineLength': self._get_avg_line_length(), \n",
    "            'stdDevLineLength': self._get_line_length_stdev(), \n",
    "        })\n",
    "\n",
    "    def _get_ast_dependent_metrics(self):\n",
    "        \"\"\"\n",
    "        Returns metrics that require importing ast module\n",
    "        \"\"\"\n",
    "        if self.tree is not None:\n",
    "            self.metrics.update({\n",
    "                'maxDecisionTokens': self._get_max_decision_tokens(), \n",
    "                'numLiteralsDensity': self._get_literals_density(),\n",
    "                'nestingDepth': self._get_max_nesting_depth(),\n",
    "                'maxDepthASTNode': self._get_max_ast_node_depth(), \n",
    "                'branchingFactor': self._get_branching_factor(),\n",
    "                'avgParams': self._get_avg_func_params(), \n",
    "                'stdDevNumParams': self._get_func_params_stdev(), \n",
    "                'avgFunctionLength': self._get_avg_function_length(), \n",
    "                'avgIdentifierLength': self._get_avg_identifier_length(), \n",
    "            })\n",
    "\n",
    "            self.metrics.update(self._get_node_type_term_frequency()) \n",
    "            self.metrics.update(self._get_node_type_avg_depth()) \n",
    "\n",
    "    def _get_tokenize_dependent_metrics(self):\n",
    "        \"\"\"\n",
    "        Returns metrics that require importing tokenize module\n",
    "        \"\"\"\n",
    "        if self.tokens is not None:\n",
    "            self.metrics.update({'numKeywordsDensity': self._get_num_keywords_density()})\n",
    "            self.metrics.update(self._get_keywords_density())\n",
    "\n",
    "    def _get_radon_dependent_metrics(self):\n",
    "        \"\"\"\n",
    "        Returns metrics that require importing radon library\n",
    "        \"\"\"\n",
    "        if self.radon_raw_metrics is not None:\n",
    "            self.metrics.update({ \n",
    "                'sloc': self.radon_raw_metrics.sloc, \n",
    "                'numVariablesDensity': self._get_num_variables_density(), \n",
    "                'numFunctionsDensity': self._get_functions_density(), \n",
    "                'numInputStmtsDensity': self._get_input_statements_density(), \n",
    "                'numAssignmentStmtDensity': self._get_assignment_statements_density(), \n",
    "                'numFunctionCallsDensity': self._get_function_calls_density(), \n",
    "                'numStatementsDensity': self._get_num_statements_density(), \n",
    "                'numClassesDensity': self._get_num_classes_density(),  \n",
    "                'cyclomaticComplexity': self._get_radon_cyclomatic_complexity(), \n",
    "                'maintainabilityIndex': self._get_radon_maintainability_index(),\n",
    "            })\n",
    "            \n",
    "    \n",
    "    def _get_ast(self):\n",
    "        \"\"\"Returns the ast of the code\"\"\"\n",
    "        try:\n",
    "            return ast.parse(self.code)\n",
    "        except SyntaxError:\n",
    "            Logger.error_log(\"Failed to parse code to ast. Returning None for ast\")\n",
    "            return None\n",
    "        \n",
    "    def _get_tokens(self):\n",
    "        \"\"\"Returns the tokens of the code using the tokenize module\"\"\"\n",
    "        try:\n",
    "            return tokenize.tokenize(BytesIO(self.code.encode('utf-8')).readline)\n",
    "        except tokenize.TokenError:\n",
    "            Logger.error_log(\"Failed to tokenize code. Returning None for tokens\")\n",
    "            return None\n",
    "    \n",
    "    def _get_keywords(self):\n",
    "        \"\"\" Returns the occurrences of each keyword in the code \"\"\"\n",
    "        if self.tokens is None:\n",
    "            Logger.error_log(\"Failed to tokenize code. Returning empty dictionary for keywords\")\n",
    "            return None\n",
    "        keywords = {}\n",
    "        tokens = self._get_tokens()\n",
    "        comment_state = False\n",
    "        try:\n",
    "            for token in tokens:\n",
    "                if token.type == tokenize.COMMENT:\n",
    "                    comment_state = True\n",
    "                elif token.type == tokenize.NL or token.type == tokenize.NEWLINE:\n",
    "                    comment_state = False\n",
    "                elif token.string in keyword.kwlist and not comment_state:\n",
    "                    if token.string not in keywords:\n",
    "                        keywords[token.string] = 1\n",
    "                    else:\n",
    "                        keywords[token.string] += 1\n",
    "        except:\n",
    "            Logger.error_log(\"Failed to tokenize code. Returning empty dictionary for keywords\")\n",
    "            return {}\n",
    "        return keywords\n",
    "    \n",
    "    def _get_radon_metrics(self):\n",
    "        \"\"\" Returns the raw metrics calculated by Radon \"\"\"\n",
    "        try:\n",
    "            return analyze(self.code)\n",
    "        except:\n",
    "            Logger.error_log(\"Failed to calculate Radon metrics. Returning None\")\n",
    "            return None\n",
    "\n",
    "    def _get_line_length(self):\n",
    "        \"\"\" Returns the length of characters in each line of the code \"\"\"\n",
    "        lines = self.code.split('\\n')\n",
    "        # lines = code.split('\\n')\n",
    "        lines_length = [len(line) for line in lines]\n",
    "        return lines_length\n",
    "    \n",
    "    def _get_num_whitespaces(self):\n",
    "        \"\"\" Returns the total occurrences of spaces (spaces, newlines) in the code \"\"\"\n",
    "        return sum(1 for char in self.code if char.isspace())\n",
    "    \n",
    "    def _get_empty_lines(self):\n",
    "        \"\"\" Returns the total occurrences of empty lines in the code \"\"\"\n",
    "        return sum(1 for line in self.code.splitlines() if line.strip() == \"\")\n",
    "    \n",
    "    def _get_avg_line_length(self):\n",
    "        \"\"\" Returns the average length of characters in each line of code \"\"\"\n",
    "        if len(self.line_length) == 0:\n",
    "            Logger.error_log(f\"No code lines. Returning default value: {default_values[0]} for avg_line_length\")\n",
    "            return default_values[0]\n",
    "        else:\n",
    "            return round(sum(self.line_length) / len(self.line_length), 2)\n",
    "    \n",
    "    def _get_line_length_stdev(self):\n",
    "        \"\"\" Returns the standard deviation of the length of characters in each line of the code\"\"\"\n",
    "        return round(statistics.stdev(self.line_length), 2) if len(self.line_length) > 1 else 0\n",
    "    \n",
    "    def _get_max_decision_tokens(self):\n",
    "        \"\"\" Returns the maximum number of tokens in a decision path \"\"\"\n",
    "        decision_path_tokens = []\n",
    "        max_decision_tokens = 0\n",
    "\n",
    "        try:\n",
    "            for node in ast.walk(self.tree):\n",
    "                if isinstance(node, (ast.If, ast.For, ast.While)):\n",
    "                    if isinstance(node, ast.If) or isinstance(node, ast.While):\n",
    "                        condition = ast.get_source_segment(self.code, node.test) \n",
    "                    elif isinstance(node, ast.For):\n",
    "                        condition = ast.get_source_segment(self.code, node)\n",
    "                    tokens = self._get_for_loop_tokens(condition)\n",
    "                    decision_path_tokens.append(tokens)\n",
    "            if decision_path_tokens:\n",
    "                max_decision_tokens = max(len(tokens) for tokens in decision_path_tokens)\n",
    "            return max_decision_tokens\n",
    "        except Exception as e:\n",
    "            Logger.error_log(f\"Failed to get max_decision_tokens: {e}. Returning default value: {default_values[0]} for max_decision_tokens\")\n",
    "            return default_values[0]\n",
    "        \n",
    "    def _get_for_loop_tokens(self, condition):\n",
    "        split = 1\n",
    "        while True:\n",
    "            try:\n",
    "                split_parts = condition.split(':', split)\n",
    "                if split == 1:\n",
    "                    condition_split = split_parts[0]\n",
    "                else:\n",
    "                    condition_split = \":\".join(split_parts[:split])\n",
    "                tokens = tokenize.tokenize(io.BytesIO(condition_split.encode('utf-8')).readline)\n",
    "                tokens = [token.string.strip() for token in tokens if token.string.strip() and token.string.strip() not in ('if', 'while', 'for', 'utf-8')]\n",
    "                return tokens\n",
    "            except tokenize.TokenError:\n",
    "                if \":\" in condition[split + 1:]:\n",
    "                    split += 1\n",
    "                else:\n",
    "                    return []\n",
    "    \n",
    "    def _get_literals_density(self):\n",
    "        \"\"\"\n",
    "        Calculates the logarithm of the total occurrences of literals \n",
    "        divided by the length of the code in terms of characters.\n",
    "\n",
    "        Returns:\n",
    "            float: The logarithm of the total occurrences of literals divided by the length of the code\n",
    "        \"\"\"\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for literals_density\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for literals_density\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for literals_density\")\n",
    "            return default_values[0]\n",
    "        literals_sum = sum(1 for node in ast.walk(self.tree) if isinstance(node, ast.Constant))\n",
    "        sum_div_by_length = literals_sum / self.radon_raw_metrics.sloc\n",
    "        return round(sum_div_by_length, 2)\n",
    "    \n",
    "    def _get_num_functions(self):\n",
    "        \"\"\" Returns the total number of functions in the code \"\"\"\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for num_functions\")\n",
    "            return default_values[0]\n",
    "        return sum(isinstance(node, ast.FunctionDef) for node in ast.walk(self.tree))\n",
    "    \n",
    "    def _get_functions_params(self):\n",
    "        \"\"\" Returns the parameters of each function in the code \"\"\"\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for functions_params\")\n",
    "            return default_values[0]\n",
    "        arguments_per_function = []\n",
    "        for node in ast.walk(self.tree): \n",
    "            if isinstance(node, ast.FunctionDef):\n",
    "                arguments_per_function.append(len(node.args.args)) \n",
    "        return arguments_per_function\n",
    "    \n",
    "    def _get_max_nesting_depth(self):\n",
    "        \"\"\" Returns the maximum nesting depth of loops, conditional statements, and functions in the code\"\"\"\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for max_nesting_depth\")\n",
    "            return default_values[0]\n",
    "        max_nesting_depth = 0\n",
    "        queue = deque([(self.tree, 0)])\n",
    "        while queue:\n",
    "            node, depth = queue.popleft()            \n",
    "            if isinstance(node, (ast.If, ast.While, ast.For, ast.FunctionDef)):\n",
    "                max_nesting_depth = max(max_nesting_depth, depth)            \n",
    "            for child_node in ast.iter_child_nodes(node):\n",
    "                queue.append((child_node, depth + 1))\n",
    "        return max_nesting_depth\n",
    "    \n",
    "    def _get_max_nesting_function(self):\n",
    "        \"\"\"Returns the maximum nesting depth of functions in the code\"\"\"\n",
    "        max_nesting_function = 0\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for max_nesting_function\")\n",
    "            return default_values[0]\n",
    "        queue = deque([(self.tree, 0)])\n",
    "        while queue:\n",
    "            node, depth = queue.popleft()\n",
    "            if isinstance(node, ast.FunctionDef):\n",
    "                max_nesting_function = max(max_nesting_function, depth)\n",
    "            for child_node in ast.iter_child_nodes(node):\n",
    "                queue.append((child_node, depth + 1))\n",
    "        return max_nesting_function\n",
    "    \n",
    "    def _get_branching_factor(self):\n",
    "        \"\"\"\n",
    "        Returns the average branching factor of the AST. \n",
    "        The branching factor is the number of children of each parent node.\n",
    "        The average branching factor is the total number of branches divided by the total number of parent nodes.\n",
    "        \"\"\"\n",
    "        #use a queue for BFS\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for branching_factor\")\n",
    "            return default_values[0]\n",
    "        queue = deque([self.tree])\n",
    "        branches = []\n",
    "\n",
    "        while queue:\n",
    "            current_node = queue.popleft()\n",
    "            current_node_branches = sum(1 for _ in ast.iter_child_nodes(current_node))\n",
    "            if current_node_branches > 0:\n",
    "                branches.append(current_node_branches)\n",
    "            #add children to the queue for further processing\n",
    "            queue.extend(ast.iter_child_nodes(current_node))\n",
    "\n",
    "        total_branches = sum(branches)\n",
    "        total_parent_nodes = len(branches)\n",
    "        average_branching_factor = round((total_branches / total_parent_nodes), 2) if total_parent_nodes != 0 else 0\n",
    "        return average_branching_factor\n",
    "    \n",
    "    def _get_avg_func_params(self):\n",
    "        \"\"\"Returns the average number of parameters of each function in the code\"\"\"\n",
    "        if self.functions_params == default_values[0]:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for avg_func_params\")\n",
    "            return default_values[0]\n",
    "        return round(sum(self.functions_params) / len(self.functions_params), 2) if len(self.functions_params) > 0 else 0.0\n",
    "    \n",
    "    def _get_func_params_stdev(self):\n",
    "        \"\"\" Returns the standard deviation of the parameters of each function in the code\"\"\"\n",
    "        if self.functions_params == default_values[0]:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for func_params_stdev\")\n",
    "            return default_values[0]\n",
    "        return round(statistics.stdev(self.functions_params), 2) if len(self.functions_params) > 1 else 0\n",
    "    \n",
    "    def _get_max_ast_node_depth(self, node=None, max_depth=0):\n",
    "        \"\"\"Returns the maximum depth (distance from root to leaf nodes) of the AST\"\"\"\n",
    "        max_depth = 0\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for max_ast_node_depth\")\n",
    "            return default_values[0]\n",
    "        queue = deque([(self.tree, 0)])\n",
    "        while queue:\n",
    "            current_node, depth = queue.popleft()\n",
    "            max_depth = max(max_depth, depth)\n",
    "            for child_node in ast.iter_child_nodes(current_node):\n",
    "                queue.append((child_node, depth + 1))\n",
    "        return max_depth\n",
    "    \n",
    "    def _get_input_statements_density(self):\n",
    "        \"\"\" Returns the total number of input statements divided by source code lines \"\"\"\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for num_input_statements\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for num_input_statements\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for num_input_statements\")\n",
    "            return default_values[0]\n",
    "        num_input_statements = sum(isinstance(node, ast.Call) and hasattr(node.func, 'id') and node.func.id == 'input' for node in ast.walk(self.tree))\n",
    "        return round((num_input_statements / self.radon_raw_metrics.sloc), 2)\n",
    "    \n",
    "    def _get_assignment_statements_density(self):\n",
    "        \"\"\" Returns the total number of assignment statements divided by source code lines \"\"\"\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for num_assignment_statements\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for num_assignment_statements\")\n",
    "            return default_values[0]\n",
    "        assignment_statements = sum(isinstance(node, ast.Assign) for node in ast.walk(self.tree))\n",
    "        return round((assignment_statements / self.radon_raw_metrics.sloc), 2)\n",
    "    \n",
    "    def _get_function_calls_density(self):\n",
    "        \"\"\" Returns the total number of function calls divided by source code lines \"\"\"\n",
    "        function_calls = sum(isinstance(node, ast.Call) for node in ast.walk(self.tree))\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for num_function_calls\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for num_function_calls\")\n",
    "            return default_values[0]\n",
    "        return round((function_calls / self.radon_raw_metrics.sloc), 2)\n",
    "    \n",
    "    def _get_num_statements_density(self):\n",
    "        \"\"\" Returns the total number of statements divided by source code lines \"\"\"\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for num_statements\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for num_statements\")\n",
    "            return default_values[0]\n",
    "        num_statements = sum(isinstance(node, ast.stmt) for node in ast.walk(self.tree))\n",
    "        return round((num_statements / self.radon_raw_metrics.sloc), 2)\n",
    "    \n",
    "    def _get_avg_function_length(self):\n",
    "        \"\"\" Returns the average length of functions in the code \"\"\"\n",
    "        function_lengths = []\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for avg_function_length\")\n",
    "            return default_values[0]\n",
    "        for node in ast.walk(self.tree):\n",
    "            if isinstance(node, ast.FunctionDef):\n",
    "                function_length = node.end_lineno - node.lineno\n",
    "                function_lengths.append(function_length)\n",
    "        return round(sum(function_lengths) / len(function_lengths), 2) if len(function_lengths) > 0 else 0.0\n",
    "    \n",
    "    def _get_avg_identifier_length(self):\n",
    "        \"\"\" Returns the average length of identifiers in the code \"\"\"\n",
    "        identifiers = set()\n",
    "        lengths = []\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for avg_identifier_length\")\n",
    "            return default_values[0]\n",
    "        for node in ast.walk(self.tree):\n",
    "            if isinstance(node, ast.Name):\n",
    "                identifiers.add(node.id)\n",
    "        lengths = [len(identifier) for identifier in identifiers]\n",
    "        return round(sum(lengths) / len(lengths), 2) if len(lengths) > 0 else 0.0\n",
    "    \n",
    "    def _get_num_classes_density(self):\n",
    "        \"\"\" Returns the total number of classes divided by source code lines \"\"\"\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for num_classes_density\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for num_classes_density\")\n",
    "            return default_values[0]\n",
    "        num_classes = sum(isinstance(node, ast.ClassDef) for node in ast.walk(self.tree))\n",
    "        return round(num_classes / self.radon_raw_metrics.sloc, 2)\n",
    "    \n",
    "    def _get_node_type_term_frequency(self):\n",
    "        \"\"\"\n",
    "        Returns the term frequency (frequency of node type with respect to the total number of nodes)\n",
    "        of each node type in the AST excluding leaves\n",
    "        \"\"\"\n",
    "        term_frequency = {}\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning empty dictionary for node_type_term_frequency\")\n",
    "            return term_frequency\n",
    "        \n",
    "        node_types = [node.__class__.__name__ for node in ast.walk(self.tree) if list(ast.iter_child_nodes(node))]\n",
    "        frequency = Counter(node_types)\n",
    "        term_frequency = {'nttf_' + k : v for k,v in frequency.items()}\n",
    "        return term_frequency\n",
    "    \n",
    "    def _get_node_type_avg_depth(self):\n",
    "        \"\"\"Returns the average depth of each node type excluding leaves in the AST\"\"\"\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning empty dictionary for node_type_avg_depth\")\n",
    "            return {}\n",
    "\n",
    "        node_queue = deque([(self.tree, 0)])\n",
    "        depth_dict = {}\n",
    "\n",
    "        while node_queue:\n",
    "            current_node, depth = node_queue.popleft()\n",
    "            node_type = type(current_node).__name__\n",
    "            if node_type not in depth_dict:\n",
    "                depth_dict[node_type] = []\n",
    "            depth_dict[node_type].append(depth) \n",
    "            # add child nodes to the queue for further processing\n",
    "            for child_node in ast.iter_child_nodes(current_node):\n",
    "                if list(ast.iter_child_nodes(child_node)): #exclude child leaves\n",
    "                    node_queue.append((child_node, depth + 1))\n",
    "\n",
    "        average_type_depths = {\n",
    "            node_type: round(sum(depths) / len(depths), 2)\n",
    "            for node_type, depths in depth_dict.items()\n",
    "        }\n",
    "\n",
    "        average_type_depths = {'ntad_' + k: v for k,v in average_type_depths.items()}\n",
    "        return average_type_depths\n",
    "    \n",
    "    def _get_num_keywords_density(self):\n",
    "        \"\"\" Returns the log of the total occurrences of keywords in the code \n",
    "        divided by the length of the code in terms of characters. \"\"\"\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for keywords_density\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for keywords_density\")\n",
    "            return default_values[0]\n",
    "        keywords_sum = sum(self.keywords.values())\n",
    "        sum_div_by_length = keywords_sum / self.radon_raw_metrics.sloc\n",
    "        return round(sum_div_by_length, 2)\n",
    "\n",
    "    def _get_keywords_density(self):\n",
    "        \"\"\"Returns the count of each keyword in the code divided by source code lines\"\"\"\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning empty dictionary for keywords_density\")\n",
    "            return {}\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning empty dictionary for keywords_density\")\n",
    "            return {}\n",
    "        keywords_density = {k: round((v / self.radon_raw_metrics.sloc), 2) for k,v in self.keywords.items()}\n",
    "        # sort the dictionary by key alphabetically\n",
    "        keywords_density = {k + \"_Density\": v for k, v in sorted(keywords_density.items(), key=lambda item: item[0])}\n",
    "        return keywords_density\n",
    "    \n",
    "    def _get_radon_halsted_metrics(self):\n",
    "        \"\"\" Returns the Halsted metrics calculated by Radon \"\"\"\n",
    "        try:\n",
    "            h_metrics = h_visit(self.code)\n",
    "            return {\n",
    "                'numberOfDistinctOperators': h_metrics.total.h1, \n",
    "                'numberOfDistinctOperands': h_metrics.total.h2, \n",
    "                'totalNumberOfOperators': h_metrics.total.N1, \n",
    "                'totalNumberOfOperands': h_metrics.total.N2,\n",
    "            }\n",
    "        except:\n",
    "            Logger.error_log(\"Failed to calculate Halsted metrics. Returning default values for all (12) Halsted metrics\")\n",
    "            return {\n",
    "                'numberOfDistinctOperators': default_values[0],\n",
    "                'numberOfDistinctOperands ': default_values[0],\n",
    "                'totalNumberOfOperators': default_values[0],\n",
    "                'totalNumberOfOperands': default_values[0],\n",
    "            }\n",
    "\n",
    "    def _get_radon_cyclomatic_complexity(self):\n",
    "        \"\"\" Returns the Cyclomatic Complexity calculated by Radon \"\"\"\n",
    "        try:\n",
    "            cc = ComplexityVisitor.from_code(self.code)\n",
    "            return cc.total_complexity\n",
    "        except:\n",
    "            Logger.error_log(f\"Failed to calculate Cyclomatic Complexity. Returning default value: {default_values[0]} for cyclomatic_complexity\")\n",
    "            return default_values[0]   \n",
    "\n",
    "    def _get_radon_maintainability_index(self):\n",
    "        \"\"\" Returns the Maintainability Index calculated by Radon \"\"\"\n",
    "        try:\n",
    "            mi = mi_visit(self.code, False)\n",
    "            return mi\n",
    "        except:\n",
    "            Logger.error_log(f\"Failed to calculate Maintainability Index. Returning default value: {default_values[0]} for maintainability_index\")\n",
    "            return default_values[0]\n",
    "    \n",
    "    def _get_functions_density(self):\n",
    "        \"\"\" \n",
    "        Returns the total number of functions divided by the source code lines\n",
    "        \"\"\"\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate functions density. Returning default value: {default_values[0]} for functions_density\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"sloc is 0. Returning default value: {default_values[0]} for functions_density\")\n",
    "            return default_values[0]\n",
    "        return round(self.num_functions / (self.radon_raw_metrics.sloc), 2)    \n",
    "    \n",
    "    def _get_num_variables_density(self):\n",
    "        \"\"\" Returns the total number of assignment variables divided by source code lines \"\"\"\n",
    "        variables = set()\n",
    "        if self.tree is None:\n",
    "            Logger.error_log(f\"Failed to parse code to ast. Returning default value: {default_values[0]} for num_variables\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics is None:\n",
    "            Logger.error_log(f\"Failed to calculate Radon metrics. Returning default value: {default_values[0]} for num_variables\")\n",
    "            return default_values[0]\n",
    "        if self.radon_raw_metrics.sloc == 0:\n",
    "            Logger.error_log(f\"File is empty (code length is 0). Returning default value: {default_values[0]} for num_variables\")\n",
    "            return default_values[0]\n",
    "        for node in ast.walk(self.tree):\n",
    "            if isinstance(node, ast.Assign):\n",
    "                for target in node.targets:\n",
    "                    if isinstance(target, ast.Name):\n",
    "                        variables.add(target.id)\n",
    "        return round((len(variables) / self.radon_raw_metrics.sloc), 2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {},
   "outputs": [],
   "source": [
    "def get_uniform_metrics(metrics_list):\n",
    "    unique_metrics = set()\n",
    "    for metrics in metrics_list:\n",
    "        for metric in metrics.keys():\n",
    "            unique_metrics.add(metric)\n",
    "    \n",
    "    for metrics in metrics_list:\n",
    "        for metric in unique_metrics:\n",
    "            if metric not in metrics:\n",
    "                metrics[metric] = 0\n",
    "        file = metrics.pop('filename')\n",
    "        \n",
    "        metrics['filename'] = file\n",
    "        \n",
    "        if file.startswith('human'):\n",
    "            metrics['label'] = 0\n",
    "        elif file.startswith('ai'):\n",
    "            metrics['label'] = 1\n",
    "    return metrics_list\n",
    "\n",
    "#save metrics to csv\n",
    "def save_metrics_to_csv(metrics, file_path):\n",
    "    print(f\"Saving metrics to {file_path}\")\n",
    "    print(metrics)\n",
    "    if metrics:\n",
    "        with open(file_path, 'w') as f:\n",
    "            writer = csv.DictWriter(f, fieldnames=metrics[0].keys())\n",
    "            writer.writeheader()\n",
    "            writer.writerows(metrics)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {},
   "outputs": [],
   "source": [
    "def extract_metrics(path, output_file):\n",
    "    metrics_list = []\n",
    "    for file in os.listdir(path):\n",
    "        if file.endswith(\".py\"):\n",
    "            Logger.info_log(f\"Extracting metrics for: {file}\")\n",
    "            with open(os.path.join(path, file), \"r\") as f:\n",
    "                code = f.read()\n",
    "            extracted_metrics = CodeMetricsExtractor(code)\n",
    "            metrics = extracted_metrics.metrics  \n",
    "            Logger.info_log(f\"Metrics extracted for: {file}\\n\\n\")  \n",
    "            if extracted_metrics.metrics:\n",
    "                metrics.update({'filename': file})\n",
    "                metrics_list.append(metrics)\n",
    "\n",
    "    uniform_metrics = get_uniform_metrics(metrics_list)\n",
    "    save_metrics_to_csv(uniform_metrics, output_file)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "input_path =  \"path/to/unverified/\"\n",
    "output_file = 'path/to/data/features/non_gameable_features.csv'\n",
    "extract_metrics(input_path, output_file)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "research",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.2"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
